﻿#include "osecpu.h"


#if (DRV_OSNUM == 0x0002)
//
// for Mac OSX 32-bit
//
#include <mach/mach.h>
#include <Cocoa/Cocoa.h>

@interface OSECPUView : NSView
{
	unsigned char *_buf;
	int _sx;
	int _sy;
	CGContextRef _context;
}

- (id)initWithFrame:(NSRect)frameRect buf:(unsigned char *)buf sx:(int)sx sy:(int)sy;
- (void)drawRect:(NSRect)rect;
@end

@implementation OSECPUView
- (id)initWithFrame:(NSRect)frameRect buf:(unsigned char *)buf sx:(int)sx sy:(int)sy
{
	self = [super initWithFrame:frameRect];
	if (self) {
		_buf = buf;
		_sx = sx;
		_sy = sy;
	}
	return self;
}

- (void)drawRect:(NSRect)rect {
	CGColorSpaceRef colorSpace;
	CGImageRef image;
	CGContextRef currentContext;
	
	currentContext = [[NSGraphicsContext currentContext] graphicsPort];
	colorSpace= CGColorSpaceCreateDeviceRGB();
	
	// ビットマップコンテキストにvramイメージを生成
	_context = CGBitmapContextCreate (_buf, _sx, _sy, 8, 4 * _sx, colorSpace, (kCGBitmapByteOrder32Little | kCGImageAlphaNoneSkipFirst));
	image = CGBitmapContextCreateImage(_context);
	
	// 生成したvramイメージをViewに描画
	CGContextDrawImage(currentContext, NSRectToCGRect(rect), image);
	
	// メモリ解放
	CFRelease(colorSpace);
	CFRelease(image);
}

- (void)keyDown:(NSEvent *)theEvent
{
	int i, j;
	unichar kc;
	
	if([[theEvent characters] length] != 1){
		return;
	}
	
	kc = [[theEvent characters] characterAtIndex:0];
	
	// 通常キー
	i = KEY_EMPTY;
	if (kc == '\r')	i = KEY_ENTER;
	if (kc == 0x1b)	i = KEY_ESC;
	if (kc == 0x7F)		i = KEY_BACKSPACE;
	if (kc == '\t')		i = KEY_TAB;
	if (kc == NSPageUpFunctionKey)		i = KEY_PAGEUP;
	if (kc == NSPageDownFunctionKey)		i = KEY_PAGEDWN;
	if (kc == NSEndFunctionKey)		i = KEY_END;
	if (kc == NSHomeFunctionKey)		i = KEY_HOME;
	if (kc == NSLeftArrowFunctionKey)		i = KEY_LEFT;
	if (kc == NSRightArrowFunctionKey)		i = KEY_RIGHT;
	if (kc == NSUpArrowFunctionKey)		i = KEY_UP;
	if (kc == NSDownArrowFunctionKey)		i = KEY_DOWN;
	if (kc == NSInsertFunctionKey)	i = KEY_INS;
	if (kc == NSDeleteFunctionKey)	i = KEY_DEL;
	if (i == KEY_EMPTY && isascii(kc)){
		i = kc;
	}
	// 修飾キー
	// TODO : 下の方の右辺、iとjが混ざっているのは修正すべき？
	j = 0;
	/*
	if ((GetKeyState(VK_LCONTROL) & (1 << 15)) != 0) j |= KEY_MODIFIER_CONTROL_L;
	if ((GetKeyState(VK_LMENU)    & (1 << 15)) != 0) j |= KEY_MODIFIER_MENU_L;
	if ((GetKeyState(VK_RCONTROL) & (1 << 15)) != 0) j |= KEY_MODIFIER_CONTROL_R;
	if ((GetKeyState(VK_RMENU)    & (1 << 15)) != 0) j |= KEY_MODIFIER_MENU_R;
	if ((GetKeyState(VK_RSHIFT)   & (1 << 15)) != 0) i |= KEY_MODIFIER_SHIFT_R;
	if ((GetKeyState(VK_LSHIFT)   & (1 << 15)) != 0) i |= KEY_MODIFIER_SHIFT_L;
	if ((GetKeyState(VK_NUMLOCK)  & (1 << 0)) != 0) i |= KEY_MODIFIER_NUMLOCK;
	if ((GetKeyState(VK_CAPITAL)  & (1 << 0)) != 0) i |= KEY_MODIFIER_CAPITAL;
	 */
	
	// NSLog(@"%X", i);
	if (i != -1) {
		putKeybuf(i | j);
	}
}
/*
- (void)keyUp:(NSEvent *)theEvent
{
	NSLog(@"kUp");
}
 */
@end

@interface Main : NSObject<NSWindowDelegate>
{
	int argc;
	const unsigned char **argv;
	char *winClosed;
	OSECPUView *_view;
	NSWindow *mainWindow;
	NSMenu *mainMenu;
}

- (void)runApp;
- (void)createThread:(int)_argc args:(const unsigned char **)_argv;
- (BOOL)windowShouldClose:(id)sender;
- (void)openWin:(unsigned char *)buf sx:(int)sx sy:(int)sy winClosed:(char *)_winClosed;
- (void)flushWin:(NSRect)rect;
@end

@implementation Main
- (void)runApp
{
	// Sub thread
	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init];
	int retv;
	retv = HeavyOSECPUMain(argc, (char **)argv);
	[NSApp terminate:self];
	[pool release];
	exit(retv);
}

- (void)createThread : (int)_argc args:(const unsigned char **)_argv
{
	argc = _argc;
	argv = _argv;
	NSThread *thread = [[[NSThread alloc] initWithTarget:self selector:@selector(runApp) object:nil] autorelease];
	[thread start];
}

- (BOOL)windowShouldClose:(id)sender
{
	*winClosed = 1;
	return YES;
}

- (void)windowDidBecomeKey:(NSNotification *)notification
{
	[mainWindow makeFirstResponder:_view];
}

- (void)openWin:(unsigned char *)buf sx:(int)sx sy:(int) sy winClosed:(char *)_winClosed
{
	mainWindow = [[NSWindow alloc] initWithContentRect: NSMakeRect(0, 0, sx, sy) styleMask: NSTitledWindowMask | NSMiniaturizableWindowMask | NSClosableWindowMask backing: NSBackingStoreBuffered defer: NO];
	mainWindow.delegate = self;
	[mainWindow setTitle: @"osecpu"];
	[mainWindow center];
	[mainWindow makeKeyAndOrderFront:self];
	[mainWindow setReleasedWhenClosed:YES];
	
	winClosed = _winClosed;
	
	_view = [[OSECPUView alloc] initWithFrame:NSMakeRect(0,0,sx,sy) buf:buf sx:sx sy:sy];
	[mainWindow.contentView addSubview:_view];
}

- (void)flushWin : (NSRect)rect
{
	[_view setNeedsDisplayInRect:rect];
}

@end

id objc_main;

int main(int argc, char **argv)
{
	// Main thread
	NSAutoreleasePool* pool;
	ProcessSerialNumber psn = {0, kCurrentProcess};
	
	// ForegroundApplicationになれるように設定(こうしないとキーイベントが取得できない)
	TransformProcessType(&psn, kProcessTransformToForegroundApplication);
		
	objc_main = [[Main alloc] init];
	NSApp = [NSApplication sharedApplication];
	
	pool = [[NSAutoreleasePool alloc] init];
	[objc_main createThread:argc args:(const unsigned char **)argv];
	[NSApp run];
	[pool release];
	return 0;
}

void *mallocRWE(int bytes)
{
	void *p = malloc(bytes);
	vm_protect(mach_task_self(), (vm_address_t) p, bytes, FALSE, VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE);
	return p;
}

void drv_openWin(int sx, int sy, unsigned char *buf, char *winClosed)
{
	[objc_main openWin:buf sx:sx sy:sy winClosed:winClosed];
}

void drv_flshWin(int sx, int sy, int x0, int y0)
{
	[objc_main flushWin:NSMakeRect(x0,y0,sx,sy)];
}

void drv_sleep(int msec)
{
	[NSThread sleepForTimeInterval:0.001*msec];
	return;
}


#elif (DRV_OSNUM == 0x0003)

#error "Your OS is not supported."

#elif (DRV_OSNUM == 0x0001)
//
// for Windows 32-bit
//
#include <windows.h>

#define TIMER_ID			 1
#define TIMER_INTERVAL		10

struct BLD_WORK {
	HINSTANCE hi;
	HWND hw;
	BITMAPINFO bmi;
	int tmcount1, tmcount2, flags, smp; /* bit0: 終了 */
	HANDLE mtx;
	char *winClosed;
};

struct BLD_WORK bld_work;

struct BL_WIN {
	int xsiz, ysiz, *buf;
};

struct BL_WORK {
	struct BL_WIN win;
	jmp_buf jb;
	int csiz_x, csiz_y, cx, cy, col0, col1, tabsiz, slctwin;
	int tmcount, tmcount0, mod, rand_seed;
	int *cbuf;
	unsigned char *ftyp;
	unsigned char **fptn;
	int *ccol, *cbak;
	int *kbuf, kbuf_rp, kbuf_wp, kbuf_c;
};

struct BL_WORK bl_work;

#define BL_SIZ_KBUF			8192

#define BL_WAITKEYF		0x00000001
#define BL_WAITKEYNF	0x00000002
#define BL_WAITKEY		0x00000003
#define BL_GETKEY		0x00000004
#define BL_CLEARREP		0x00000008
#define BL_DELFFF		0x00000010

#define	BL_KEYMODE		0x00000000	// 作りかけ, make/remake/breakが見えるかどうか

#define w	bl_work
#define dw	bld_work

void bld_openWin(int x, int y, char *winClosed);
void bld_flshWin(int sx, int sy, int x0, int y0);
LRESULT CALLBACK WndProc(HWND hw, unsigned int msg, WPARAM wp, LPARAM lp);
void bl_cls();
int bl_iCol(int i);
void bl_readyWin(int n);

static HANDLE threadhandle;

int main(int argc, char **argv)
{
	// Program entry point
	return HeavyOSECPUMain(argc, argv);
}

void *mallocRWE(int bytes)
{
	void *p = malloc(bytes);
	DWORD dmy;
	VirtualProtect(p, bytes, PAGE_EXECUTE_READWRITE, &dmy);
	return p;
}

static int winthread(void *dmy)
{
	WNDCLASSEX wc;
	RECT r;
	int i, x, y;
	MSG msg;
	
	x = dw.bmi.bmiHeader.biWidth;
	y = -dw.bmi.bmiHeader.biHeight;
	
	wc.cbSize = sizeof (WNDCLASSEX);
	wc.style = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc = WndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = dw.hi;
	wc.hIcon = (HICON)LoadImage(NULL, MAKEINTRESOURCE(IDI_APPLICATION),
								IMAGE_ICON, 0, 0, LR_DEFAULTSIZE | LR_SHARED);
	wc.hIconSm = wc.hIcon;
	wc.hCursor = (HCURSOR)LoadImage(NULL, MAKEINTRESOURCE(IDC_ARROW),
									IMAGE_CURSOR, 0, 0, LR_DEFAULTSIZE | LR_SHARED);
	wc.hbrBackground = (HBRUSH)COLOR_APPWORKSPACE;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = "WinClass";
	if (RegisterClassEx(&wc) == 0)
		return 1;
	r.left = 0;
	r.top = 0;
	r.right = x;
	r.bottom = y;
	AdjustWindowRect(&r, WS_OVERLAPPEDWINDOW, FALSE);
	x = r.right - r.left;
	y = r.bottom - r.top;
	
	char *t = "osecpu";
	
	dw.hw = CreateWindowA("WinClass", t, WS_OVERLAPPEDWINDOW,
						  CW_USEDEFAULT, CW_USEDEFAULT, x, y, NULL, NULL, dw.hi, NULL);
	if (dw.hw == NULL)
		return 1;
	ShowWindow(dw.hw, SW_SHOW);
	UpdateWindow(dw.hw);
	SetTimer(dw.hw, TIMER_ID, TIMER_INTERVAL, NULL);
	SetTimer(dw.hw, TIMER_ID + 1, TIMER_INTERVAL * 10, NULL);
	SetTimer(dw.hw, TIMER_ID + 2, TIMER_INTERVAL * 100, NULL);
	dw.flags |= 2 | 4;
	
	for (;;) {
		i = GetMessage(&msg, NULL, 0, 0);
		if (i == 0 || i == -1)	/* エラーもしくは終了メッセージ */
			break;
		/* そのほかはとりあえずデフォルト処理で */
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	//	PostQuitMessage(0);
	dw.flags |= 1; /* 終了, bld_waitNF()が見つける */
	if (dw.winClosed != NULL)
		*dw.winClosed = 1;
	return 0;
}

void bld_openWin(int sx, int sy, char *winClosed)
{
	static int i;
	
	dw.bmi.bmiHeader.biSize = sizeof (BITMAPINFOHEADER);
	dw.bmi.bmiHeader.biWidth = sx;
	dw.bmi.bmiHeader.biHeight = -sy;
	dw.bmi.bmiHeader.biPlanes = 1;
	dw.bmi.bmiHeader.biBitCount = 32;
	dw.bmi.bmiHeader.biCompression = BI_RGB;
	dw.winClosed = winClosed;
	
	threadhandle = CreateThread(NULL, 0, (void *)&winthread, NULL, 0, (void *)&i);
	
	return;
}

void drv_flshWin(int sx, int sy, int x0, int y0)
{
	InvalidateRect(dw.hw, NULL, FALSE);
	UpdateWindow(dw.hw);
	return;
}

LRESULT CALLBACK WndProc(HWND hw, unsigned int msg, WPARAM wp, LPARAM lp)
{
	int i, j;
	if (msg == WM_PAINT) {
		PAINTSTRUCT ps;
		HDC hdc = BeginPaint(dw.hw, &ps);
		SetDIBitsToDevice(hdc, 0, 0, w.win.xsiz, w.win.ysiz,
						  0, 0, 0, w.win.ysiz, w.win.buf, &dw.bmi, DIB_RGB_COLORS);
		EndPaint(dw.hw, &ps);
	}
	if (msg == WM_DESTROY) {
		PostQuitMessage(0);
		return 0;
	}
	if (msg == WM_TIMER && wp == TIMER_ID) {
		w.tmcount += TIMER_INTERVAL;
		return 0;
	}
	if (msg == WM_TIMER && wp == TIMER_ID + 1) {
		dw.tmcount1 += TIMER_INTERVAL * 10;
		w.tmcount = dw.tmcount1;
		return 0;
	}
	if (msg == WM_TIMER && wp == TIMER_ID + 2) {
		dw.tmcount2 += TIMER_INTERVAL * 100;
		w.tmcount = dw.tmcount1 = dw.tmcount2;
		return 0;
	}
	if (msg == WM_KEYDOWN || msg == WM_SYSKEYDOWN) {
		// 通常キー
		i = KEY_EMPTY;
		if (wp == VK_RETURN)	i = KEY_ENTER;
		if (wp == VK_ESCAPE)	i = KEY_ESC;
		if (wp == VK_BACK)		i = KEY_BACKSPACE;
		if (wp == VK_TAB)		i = KEY_TAB;
		if (wp == VK_PRIOR)		i = KEY_PAGEUP;
		if (wp == VK_NEXT)		i = KEY_PAGEDWN;
		if (wp == VK_END)		i = KEY_END;
		if (wp == VK_HOME)		i = KEY_HOME;
		if (wp == VK_LEFT)		i = KEY_LEFT;
		if (wp == VK_RIGHT)		i = KEY_RIGHT;
		if (wp == VK_UP)		i = KEY_UP;
		if (wp == VK_DOWN)		i = KEY_DOWN;
		if (wp == VK_INSERT)	i = KEY_INS;
		if (wp == VK_DELETE)	i = KEY_DEL;
		// 修飾キー
		// TODO : 下の方の右辺、iとjが混ざっているのは修正すべき？
		j = 0;
		if ((GetKeyState(VK_LCONTROL) & (1 << 15)) != 0) j |= KEY_MODIFIER_CONTROL_L;
		if ((GetKeyState(VK_LMENU)    & (1 << 15)) != 0) j |= KEY_MODIFIER_MENU_L;
		if ((GetKeyState(VK_RCONTROL) & (1 << 15)) != 0) j |= KEY_MODIFIER_CONTROL_R;
		if ((GetKeyState(VK_RMENU)    & (1 << 15)) != 0) j |= KEY_MODIFIER_MENU_R;
		if ((GetKeyState(VK_RSHIFT)   & (1 << 15)) != 0) i |= KEY_MODIFIER_SHIFT_R;
		if ((GetKeyState(VK_LSHIFT)   & (1 << 15)) != 0) i |= KEY_MODIFIER_SHIFT_L;
		if ((GetKeyState(VK_NUMLOCK)  & (1 << 0)) != 0) i |= KEY_MODIFIER_NUMLOCK;
		if ((GetKeyState(VK_CAPITAL)  & (1 << 0)) != 0) i |= KEY_MODIFIER_CAPITAL;
		if (j != 0) {
			if ('A' <= wp && wp <= 'Z'){
				i = wp;
			}
		}
		if (i != -1) {
			putKeybuf(i | j);
			return 0;
		}
	}
	if (msg == WM_KEYUP) {
		i = 0xfff;
	}
	if (msg == WM_CHAR) {
		i = 0;
		if (' ' <= wp && wp <= 0x7e) {
			i = wp;
			j &= 0;
			if ((GetKeyState(VK_LCONTROL) & (1 << 15)) != 0) j |= 1 << 17;
			if ((GetKeyState(VK_LMENU)    & (1 << 15)) != 0) j |= 1 << 18;
			if ((GetKeyState(VK_RCONTROL) & (1 << 15)) != 0) j |= 1 << 25;
			if ((GetKeyState(VK_RMENU)    & (1 << 15)) != 0) j |= 1 << 26;
			if ((GetKeyState(VK_RSHIFT)   & (1 << 15)) != 0) i |= 1 << 24;
			if ((GetKeyState(VK_LSHIFT)   & (1 << 15)) != 0) i |= 1 << 16;
			if ((GetKeyState(VK_NUMLOCK)  & (1 << 0)) != 0) i |= 1 << 22;
			if ((GetKeyState(VK_CAPITAL)  & (1 << 0)) != 0) i |= 1 << 23;
			if (('A' <= wp && wp <= 'Z') || ('a' <= wp && wp <= 'z')) {
				if (j != 0) {
					i |= j;
					i &= ~0x20;
				}
			}
			putKeybuf(i);
			return 0;
		}
	}
	return DefWindowProc(hw, msg, wp, lp);
}

void drv_openWin(int sx, int sy, UCHAR *buf, char *winClosed)
{
	w.win.buf = (int *)buf;
	w.win.xsiz = sx;
	w.win.ysiz = sy;
	bld_openWin(sx, sy, winClosed);
	return;
}

void drv_sleep(int msec)
{
	Sleep(msec);
	//	MsgWaitForMultipleObjects(1, &threadhandle, FALSE, msec, QS_ALLINPUT);
	/* 勉強不足でまだ書き方が分かりません! */
	return;
}

#else

#error "Illegal OS type. Edit osecpu.h and look at DRV_OSNUM"

#endif