﻿#include "osecpu.h"

HOSECPU_Device_Window mainWindow;
HOSECPU_Device_Keyboard mainKeyboard;
HOSECPU_RuntimeEnvironment *dbg_env;

unsigned char *Init_LoadSysLib(char argv0[]);
void LoadAppBin(HOSECPU_RuntimeEnvironment *env);

int HeavyOSECPUMain(int argc, char **argv)
{
	HOSECPU_RuntimeEnvironment env;
	unsigned char *pAppJITBin;
	unsigned char *appBinDevFuncBase;
	unsigned char *pSyslibJITBin;
	unsigned char *syslibJITBinBase;
	HOSECPU_LabelListTag *label;
	HOSECPU_PointerControlTag *ptrCtrl;
	// For decoder work
	unsigned char *bufP06;	// P06	UINT8	>=2M
	unsigned int *bufP0A;	// P0A	UINT32	16K
	int *pxxFlag;			// P0B	SINT32	64
	int *typLabel;			// P0C	SINT32	4K
	unsigned char *opTbl;	// P0D	UINT8	256
	// For tek decomp
	unsigned char *tekDecompTemp;
	// time counter
	double tm0, tm1, tm2;
	// options from arguments
	int argDebugMode = 0;
	int argStackSize = 1;
	//
	int i;
	const char *s;
	unsigned char *p;
	FILE *fp;
	
	//
	// Initialize
	//
	
	// For debug
	dbg_env = &env;
	// Initialize mainWindow
	mainWindow.vram = NULL;
	mainWindow.xsize = 0;
	mainWindow.ysize = 0;
	// Initialize keyInputting
	mainKeyboard.buf = malloc(KEYBUFSIZ * sizeof(int));
	mainKeyboard.r = 0;
	mainKeyboard.w = 0;
	mainKeyboard.c = 0;
	// Initialize HOSECPU_RuntimeEnvironment
	env.mainArgc = argc;
	env.mainArgv = (const char **)argv;
	//
	env.appOSECPUBin = malloc(APPBINSIZE);
	env.appOSECPUBackendBin = malloc(APPBINSIZE);
	//
	env.appJITBin = mallocRWE(APPJITSIZE);
	pAppJITBin = env.appJITBin;
	env.syslibJITBin = mallocRWE(SYSLIBJITSIZE);
	pSyslibJITBin = env.syslibJITBin;
	// syslibJITBin image:
	// env.syslibJITBin		+ - - - -
	//						| errHndl
	// syslibJITBinBase		+ - - - -
	//						| syslib
	//
	env.executionLevel = JITC_LV_SLOWEST;
	// For decoder work
	bufP06 = malloc(DECODER_BUF06_SIZE);
	bufP0A = malloc(DECODER_BUF0A_SIZE);
	pxxFlag = malloc(64 * 4);
	typLabel = malloc(4096 * 4);
	opTbl = malloc(256);
	//
	label = malloc(JITC_MAXLABELS * sizeof (HOSECPU_LabelListTag));
	ptrCtrl = malloc(PTRCTRLSIZ * sizeof (HOSECPU_PointerControlTag));
	for (i = 0; i < PTRCTRLSIZ; i++) {
		ptrCtrl[i].liveSign = 0;
		ptrCtrl[i].size = -1;
	}
	ptrCtrl[0].size = -2;
	//
	randStatInit((unsigned int)time(NULL));
	
	//
	// Load syslib
	//
	
	env.syslibOSECPUBin = Init_LoadSysLib(argv[0]);
	
	pSyslibJITBin = jitCompInit(pSyslibJITBin);
	syslibJITBinBase = pSyslibJITBin;
	// labelはjitc0()内で初期化される。
	i = jitc0(&pSyslibJITBin, env.syslibJITBin + SYSLIBJITSIZE, env.syslibOSECPUBin + 32, env.syslibOSECPUBin + SYSLIBBINSIZE, JITC_LV_SLOWEST + 9, label);
	if (i != 0){
		fputs("syslib-file JITC error.\n", stderr);
		return 1;
	}
	
	//
	// Load app
	//
	
	LoadAppBin(&env);
	
	tm0 = clock() / (double)CLOCKS_PER_SEC;
	
	if (env.appOSECPUBin[2] == 0xf0) {
		// tek5圧縮がかかっている
#if (USE_TEK5 != 0)
		tekDecompTemp = malloc(TEKDECOMPTMPSIZE);
		env.appOSECPUBinSize = tek5Decomp(env.appOSECPUBin + 2, env.appOSECPUBin + APPBINSIZE, tekDecompTemp);
		free(tekDecompTemp);
		env.appOSECPUBinSize += 2;
#else
		env.appOSECPUBinSize = -9;
#endif
		if (env.appOSECPUBinSize < 0) {
			fputs("unsupported-format(tek5)\n", stderr);
			return 1;
		}
	}
	//デバッグモード指定
	s = searchArg(argc, (const char **)argv, "debug:", 0);
	if (s != NULL){
		argDebugMode = *s - '0';
	}
	//スタックサイズ指定(MiB単位)
	s = searchArg(argc, (const char **)argv, "stack:", 0);
	if (s != NULL){
		argStackSize = strtol(s, NULL, 0);
	}
	
	/* フロントエンドコードをバックエンドコードに変換する */
	if ((env.appOSECPUBin[2] & 0xf0) != 0) {				// 3バイト目が00なら処理しない
		env.appOSECPUBackendBin[0] = env.appOSECPUBin[0];
		env.appOSECPUBackendBin[1] = env.appOSECPUBin[1];
		env.preg[2].p = &env.appOSECPUBackendBin[2];
		env.preg[3].p = &env.appOSECPUBackendBin[APPBINSIZE];
		env.preg[4].p = &env.appOSECPUBin[2];
		env.preg[5].p = &env.appOSECPUBin[env.appOSECPUBinSize];
		env.preg[6].p = bufP06;
		env.preg[7].p = bufP06 + DECODER_BUF06_SIZE;
		env.preg[0x0A].p = (unsigned char *)bufP0A;
		env.preg[0x0B].p = (void *)pxxFlag;
		env.preg[0x0C].p = (void *)typLabel;
		env.preg[0x0D].p = opTbl;
		jitcRunBinary((void *)syslibJITBinBase, &env);
		if (env.ireg[0] != 0) {
			p = env.preg[2].p - 1;
			fprintf(stderr, "unpack error: %02X (at %06X) (R00=%d)\n", *p, p - env.appOSECPUBackendBin, env.ireg[0]);
			if ((argDebugMode & 2) != 0) {
				fp = fopen("debug2.bin", "wb");
				fwrite(env.appOSECPUBackendBin, 1, p - env.appOSECPUBackendBin + 16, fp);
				fclose(fp);
			}
			exit(1);
		}
		env.appOSECPUBackendBinSize = env.preg[2].p - env.appOSECPUBackendBin;
	} else{
		memcpy(env.appOSECPUBackendBin, env.appOSECPUBin, env.appOSECPUBinSize);
		env.appOSECPUBackendBinSize = env.appOSECPUBinSize;
	}
	
	if ((argDebugMode & 2) != 0) {
		/*変換後のバックエンドコードをファイルへ保存*/
		fp = fopen("debug2.bin", "wb");
		fwrite(env.appOSECPUBackendBin, 1, env.appOSECPUBackendBinSize, fp);
		fclose(fp);
	}
	
	//JITコンパイル
	i = jitc0(&pAppJITBin, env.appJITBin + APPJITSIZE, env.appOSECPUBackendBin, env.appOSECPUBackendBin + APPBINSIZE, env.executionLevel, label);
	if (i == 1){
		fputs("app-file header error.\n", stderr);
		return 1;
	}
	if (i != 0){
		return 1;
	}
	
	env.appJITBinSize = pAppJITBin - env.appJITBin;
	
	// 末尾にAPI関数呼び出しコードを追加
	appBinDevFuncBase = pAppJITBin;
	pAppJITBin = jitCompCallFunc(pAppJITBin, &devFunc);
	
	tm1 = clock() / (double)CLOCKS_PER_SEC;
	
	/* レジスタ初期化 */
	for (i = 0; i < 64; i++){
		env.ireg[i] = 0;
	}
	for (i = 0; i < 64; i++) {
		env.preg[i].p = NULL;
		env.preg[i].typ = -1;
		env.preg[i].p0 = NULL;
		env.preg[i].p1 = NULL;
	}
	
	env.buf0 = env.buf1 = NULL;
	
	// p28にapiをコールするアドレスを設定
	env.preg[0x28].p = appBinDevFuncBase;	// appBinDevFuncBaseには、devFuncをコールするコードが書かれている
	env.preg[0x28].typ = 0; // TYP_CODE
	env.preg[0x28].p0 = appBinDevFuncBase;		// アドレス演算できる範囲を制限
	env.preg[0x28].p1 = appBinDevFuncBase + 1;	// アドレス演算できる範囲を制限
	
	//env.preg[0x00].p = malloc(1024 * 1024) + (1024 * 1024 - 32);
	env.junkStack = malloc(argStackSize << 20);
	env.junkStack1 = env.junkStack + (argStackSize << 20);
	env.winClosed = 0;
	env.autoSleep = 0;
	env.lastConsoleChar = '\n';
	
	env.label = label;
	env.maxLabels = JITC_MAXLABELS;
	env.jitbuf = pAppJITBin;
	env.jitbuf1 = env.appJITBin + APPJITSIZE;
	env.errHndl = &errorHandler;
	env.appReturnCode = 0;
	
	env.dbgr = 0;
	if (searchArg(argc, (const char **)argv, "dbgr:1", 0) != NULL){
		env.dbgr = 1;
	}
	
	if ((argDebugMode & 1) != 0) {
		fp = fopen("debug1.bin", "wb");
		fwrite(env.appJITBin, 1, pAppJITBin - env.appJITBin, fp);
		fclose(fp);
	}
	
	/* JITコード実行 */
	if (setjmp(env.setjmpEnv) == 0){
		jitcRunBinary((void *)env.appJITBin, &env);
	}
	if (env.autoSleep != 0) {
		if (mainWindow.vram != NULL){
			drv_flshWin(mainWindow.xsize, mainWindow.ysize, 0, 0);
		}
		while (env.winClosed == 0){
			drv_sleep(100);
		}
	}
	if (env.lastConsoleChar != '\n'){
		putchar('\n');
	}
	
	tm2 = clock() / (double)CLOCKS_PER_SEC;
	
	/* 実行結果確認のためのレジスタダンプ */
	if (searchArg(argc, (const char **)argv, "verbose:1", 0) != NULL) {
		printf("time: JITC=%.3f[sec], exec=%.3f[sec]\n", tm1 - tm0, tm2 - tm1);
		printf("size: OSECPU=%d, decomp=%d, native=%d\n", env.appOSECPUBinSize, env.appOSECPUBackendBinSize, env.appJITBinSize);
		printf("result:\n");
		printf("R00:0x%08X  R01:0x%08X  R02:0x%08X  R03:0x%08X\n", env.ireg[0], env.ireg[1], env.ireg[2], env.ireg[3]);
	}
#if (USE_DEBUGGER != 0)
	dbgrMain(&env);
#endif
	return env.appReturnCode;
	
	return 0;
}

unsigned char *Init_LoadSysLib(char argv0[])
{
	unsigned char *syslib;
	FILE *fp;
	unsigned char *up;
	int appsize;
	unsigned char *tmpWorkMemory;
	
	/* syslibの読み込み */
	syslib = malloc(SYSLIBBINSIZE);
	fp = fopen(SYSLIB_OSE, "rb");
	if (fp == NULL) {
		syslib[0] = '/';
		strcpy((char *)syslib + 1, argv0);
		up = syslib + 1;
		while (*up != '\0'){
			up++;
		}
		while (*up != '/' && *up != 0x5c){
			up--;
		}
		up++;
		strcpy((char *)up, SYSLIB_OSE);
		fp = fopen((char *)syslib + 1, "rb");
	}
	if (fp == NULL) {
		fputs("syslib-file fopen error.\n", stderr);
		exit(EXIT_FAILURE);
	}
	appsize = fread(syslib, 1, SYSLIBBINSIZE - 4, fp);
	fclose(fp);
	if (appsize >= SYSLIBBINSIZE - 4) {
		fputs("syslib-file too large.\n", stderr);
		exit(EXIT_FAILURE);
	}
	if (syslib[0] == 0x05 && syslib[1] == 0xc1) {
		// maklib のライブラリ形式である。
		tmpWorkMemory = malloc(SYSLIBDECOMPTMPSIZE);
		memcpy(tmpWorkMemory, syslib, appsize);
		ComLib_main(tmpWorkMemory + 2, syslib + 2);
		free(tmpWorkMemory);
		//
		syslib[0] = 0x05;
		syslib[1] = 0x1b;
	}
	
	fp = fopen("syslib_dbg.ose", "wb");
	fwrite(syslib, 1, SYSLIBBINSIZE, fp);
	fclose(fp);
	return syslib;
}

void LoadAppBin(HOSECPU_RuntimeEnvironment *env)
{
	FILE *fp;
	const char *fileName;
	/* アプリバイナリの読み込み */
	if (env->mainArgc <= 1) {
		//アプリ名未指定なので何事もなく終了
		exit(EXIT_SUCCESS);
	}
	fileName = env->mainArgv[1];
	//アプリ名先頭に:n:をつけることでレベルnでの実行が可能？
	if (fileName[0] == ':' && fileName[2] == ':') {
		env->executionLevel = fileName[1] - '0';
		if (env->executionLevel < 0 || env->executionLevel > 9){
			env->executionLevel = JITC_LV_SLOWEST;
		}
		fileName += 3;
	}
	
	fp = fopen(fileName, "rb");
	if (fp == NULL) {
		fputs("app-file load error.\n", stderr);
		exit(EXIT_FAILURE);
	}
	env->appOSECPUBinSize = fread(env->appOSECPUBin, 1, APPBINSIZE - 4, fp);
	fclose(fp);
	
	if (env->appOSECPUBinSize >= APPBINSIZE - 4) {
		fputs("app-file too large.\n", stderr);
		exit(EXIT_FAILURE);
	}
	if (env->appOSECPUBinSize < 3) {
		fputs("app-file header error.\n", stderr);
		exit(EXIT_FAILURE);
	}
}

void putKeybuf(int i)
{
	// TODO : あふれた場合のことを考慮すべき
	if (mainKeyboard.c < KEYBUFSIZ) {
		mainKeyboard.buf[mainKeyboard.w] = i;
		mainKeyboard.c++;
		mainKeyboard.w = (mainKeyboard.w + 1) & (KEYBUFSIZ - 1);
	}
	return;
}

