﻿#ifndef _HDLOAD_OSECPU
#define _HDLOAD_OSECPU 1

/* Visual Studio で fopen()やsprintf() などの古い関数を使用する時に出る警告を抑止する*/
#define _CRT_SECURE_NO_WARNINGS 1

//
// Including stdc headers
//

#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <time.h>
#include <setjmp.h>
#include <stdlib.h>


//
// Compile options
//

// Target architecture
// 1 : i386
#define JITC_ARCNUM 0x0001

// Target operating system
// 1 : Windows 32bit
// 2 : Mac OSX 32bit
// 3 : blike for Linux
#ifdef _WIN32
#define DRV_OSNUM			0x0001
#endif
#ifdef __APPLE__
#define DRV_OSNUM			0x0002
#endif
#ifdef __linux__
#define DRV_OSNUM			0x0003
#endif
//#define DRV_OSNUM 0x0002

// Decoder (syslib.ose) setting
// syslib.ose is necessary to work OSECPU
#define SYSLIB_OSE	"syslib.ose"

//
// Define constant values
//

// SIGN1: The 2nd signature of OSECPU Format(05 e1)
// It will be changed in OSECPU Rev.2 to "e2" (no adaptation in their binary layers)
#define SIGN1		0xe1

#define USE_DEBUGGER	1			// デバッグ機能を実行バイナリに含むか設定。デバッガ利用には dbgr:1 引数が実行時に必要。
#define ENABLE_DEBUG_CODE   0		// env.dbg_currentCodeに実行するコードを保存するデバッグコードを挿入する
#define ENABLE_DEBUG_CODE_ADDR	0
#define USE_TEK5		1


/* JITC mode flags */
#define	JITC_LV_SLOWEST		0	/* デバッグ支援は何でもやる */
#define	JITC_LV_SLOWER		1	/* エラーモジュールはレポートできるが、行番号は分からない、テストは過剰 */
#define	JITC_LV_SAFE		2	/* とにかく止まる、場所は不明、テストは必要最小限 */
#define	JITC_LV_FASTER		4	/* 情報は生成するがチェックをしない */
#define	JITC_LV_FASTEST		5	/* 情報すら生成しない */

#define JITC_PHASE1			0x0001
#define JITC_SKIPCHECK		0x0002	/* セキュリティチェックを省略する（高速危険モード） */
#define JITC_NOSTARTUP		0x0004

#define JITC_MAXLABELS		4096
#define PTRCTRLSIZ			4096

#define	APPBINSIZE			1 * 1024 * 1024 /* 1MB for now */
#define APPJITSIZE			1 * 1024 * 1024 /* 1MB for now */
#define	SYSLIBBINSIZE		1 * 1024 * 1024 /* 1MB for now */
#define	SYSLIBJITSIZE		2 * 1024 * 1024 /* 2MB for now */
#define SYSLIBDECOMPTMPSIZE 1 * 1024 * 1024 /* 1MB for now */
#define TEKDECOMPTMPSIZE	1 * 1024 * 1024 /* 1MB for now */
#define SYSTMP1SIZ			2 * 1024 * 1024 /* 2MB for now */

#define DECODER_BUF06_SIZE	2 * 1024 * 1024
#define DECODER_BUF0A_SIZE	1 * 1024 * 1024	// too large?

#define KEYBUFSIZ		4096

#define KEY_EMPTY       -1
#define KEY_ENTER		'\n'
#define KEY_ESC			27
#define KEY_BACKSPACE	8
#define KEY_TAB			9
#define KEY_PAGEUP		0x1020
#define KEY_PAGEDWN		0x1021
#define	KEY_END			0x1022
#define	KEY_HOME		0x1023
#define KEY_LEFT		0x1024
#define KEY_UP			0x1025
#define KEY_RIGHT		0x1026
#define KEY_DOWN		0x1027
#define KEY_INS			0x1028
#define KEY_DEL			0x1029

#define KEY_MODIFIER_SHIFT_L    (1 << 16)   // [C]
#define KEY_MODIFIER_CONTROL_L  (1 << 17)
#define KEY_MODIFIER_MENU_L     (1 << 18)
#define KEY_MODIFIER_NUMLOCK    (1 << 22)   // [C]
#define KEY_MODIFIER_CAPITAL    (1 << 23)   // [C]
#define KEY_MODIFIER_SHIFT_R    (1 << 24)   // [C]
#define KEY_MODIFIER_CONTROL_R  (1 << 25)
#define KEY_MODIFIER_MENU_R     (1 << 26)


//
// HOSECPU structures
//
typedef struct PtrCtrl HOSECPU_PointerControlTag;
struct PtrCtrl {
	int liveSign;
	int size, typ;
	unsigned char *p0;
};

typedef struct Ptr HOSECPU_PointerRegisterEntry;
struct Ptr {
	// 32バイト(=256bit!)
	unsigned char *p;
	
	/* static char *typName[] = {
		"T_CODE", "T_VPTR", "T_SINT8", "T_UINT8",
		"T_SINT16", "T_UINT16", "T_SINT32", "T_UINT32",
		"T_SINT4", "T_UINT4", "T_SINT2", "T_UINT2",
		"T_SINT1", "T_UINT1", "T_SINT12", "T_UINT12",
		"T_SINT20", "T_UINT20", "T_SINT24", "T_UINT24",
		"T_SINT28", "T_UINT28"
	} ; */
	int typ;
	unsigned char *p0, *p1;
	int liveSign;
	HOSECPU_PointerControlTag *pls;
	int flags, dummy;	/* read/writeなど */
};

typedef struct LabelTable HOSECPU_LabelListTag;
struct LabelTable {
	unsigned char *p, *p1;
	int opt;
	/*
	 * default = -1
	 * TYP_CODE = 0
	 * T_UINT8 = 3
	 *
	 * 将来的には UInt8, SInt32, Flt64, UInt8, VPtr が使えるようになる　http://osecpu.osask.jp/wiki/?page0053
	 */
	int typ;
};

typedef struct Device_Window HOSECPU_Device_Window;
struct Device_Window {
    int *vram;
    int xsize, ysize;
};

typedef struct Device_Keyboard HOSECPU_Device_Keyboard;
struct Device_Keyboard {
    int *buf;
    int r;
	int w;
	int c;
};

typedef struct Regs HOSECPU_RuntimeEnvironment;
struct Regs {
	int ireg[64];								// 整数レジスタ (4 * 64) = 256
	HOSECPU_PointerRegisterEntry preg[64];		// ポインタレジスタ (32 * 64) = 2048
    //
	int debugInfo0;								// 2304
	int debugInfo1;								// 2308
    int dbg_currentCode;                        // 2312
	int dbg_currentCodeAddress;					// 2316
	//
	HOSECPU_PointerControlTag *ptrCtrl;         // 2320
	char winClosed, autoSleep;
	jmp_buf setjmpEnv;
    int appReturnCode;      // アプリ自体の終了コード
    
    /* Main environment */
    int mainArgc;           // HOSECPU起動引数の個数
    const char **mainArgv;  // HOSECPU起動引数リスト
	//
    unsigned char *appOSECPUBin;  // 実行するアプリの元バイナリ（渡されたファイルの内容そのもの）
	int appOSECPUBinSize;
	unsigned char *appOSECPUBackendBin;  // 実行するアプリの元バイナリ（渡されたファイルの内容そのもの）
	int appOSECPUBackendBinSize;
	unsigned char *appJITBin;  // 実行するアプリのJITコンパイルされたバイナリ（実行可能）
	int appJITBinSize;
	//
    unsigned char *syslibOSECPUBin;  // syslibの元バイナリ
    unsigned char *syslibJITBin;  // syslibのJITコンパイルされたバイナリ（実行可能）
    //
    int executionLevel;
    
	/* for-junkApi */
	unsigned char *buf0, *buf1, *junkStack, lastConsoleChar, *junkStack1;
    
	HOSECPU_LabelListTag *label;
	int maxLabels;
	unsigned char *jitbuf, *jitbuf1;
	void(*errHndl)(HOSECPU_RuntimeEnvironment *);
	char dbgr;
	int mapDi1s[16][16];
};

//
// Grobal values
//

extern HOSECPU_Device_Keyboard mainKeyboard;
extern HOSECPU_Device_Window mainWindow;

//
// Functions
//

// @main.c
int HeavyOSECPUMain(int argc, char **argv);
void putKeybuf(int i);

// @comlib.c
unsigned char *ComLib_main(const unsigned char *p, unsigned char *q);
// @dpndenv.c
// OSに依存する関数群を定義する。
void *mallocRWE(int bytes); // 実行権付きメモリのmalloc.
void drv_openWin(int x, int y, unsigned char *buf, char *winClosed);
void drv_flshWin(int sx, int sy, int x0, int y0);
void drv_sleep(int msec);

// @function.c
void dbgrMain(HOSECPU_RuntimeEnvironment *r);
const char *searchArg(int argc, const char **argv, const char *tag, int i); // コマンドライン引数処理.
void devFunc(HOSECPU_RuntimeEnvironment *r); // junkApiを処理する関数

// @jitc.c
void errorHandler(HOSECPU_RuntimeEnvironment *r);
void PRegCopy(HOSECPU_PointerRegisterEntry *dst, HOSECPU_PointerRegisterEntry *src);
// @jitcx86.c
int jitc0(unsigned char **qq, unsigned char *q1, const unsigned char *p0, const unsigned char *p1, int level, HOSECPU_LabelListTag *label);
int jitCompiler(unsigned char *dst, unsigned char *dst1, const unsigned char *src, const unsigned char *src1, const unsigned char *src0, HOSECPU_LabelListTag *label, int maxLabels, int level, int debugInfo1, int flags);
unsigned char *jitCompCallFunc(unsigned char *dst, void *func);
unsigned char *jitCompInit(unsigned char *dst);
void jitcRunBinary(void (*bin)(char *), HOSECPU_RuntimeEnvironment *env);

// @randmt.c
void randStatInit(unsigned int seed);
void randStatNext();
unsigned int randGetNextUInt32(void);

// @screen.c
static int iColor1[] = {
	0x000000, 0xff0000, 0x00ff00, 0xffff00,
	0x0000ff, 0xff00ff, 0x00ffff, 0xffffff
};
void putOsaskChar(int c, HOSECPU_RuntimeEnvironment *r);
void checkString(HOSECPU_RuntimeEnvironment *r, int rxx, int pxx);
void checkRect(HOSECPU_RuntimeEnvironment *r, int rxx);
int loadColor(HOSECPU_RuntimeEnvironment *r, int rxx);

// @usetek.c
#if (USE_TEK5 != 0)
#include "tek.h"
int appackSub2(const UCHAR **pp, char *pif);
int appackSub3u(const UCHAR **pp, char *pif);
int tek5Decomp(UCHAR *buf, UCHAR *buf1, UCHAR *tmp);
#endif


#endif
